"""A dialog for viewing the text of a PEP."""

##############################################################################
# Python imports.
from functools import partial
from pathlib import Path

##############################################################################
# Textual imports.
from textual import on, work
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, TextArea

##############################################################################
# Textual enhanced imports.
from textual_enhanced.dialogs import Confirm
from textual_enhanced.tools import add_key
from textual_enhanced.widgets import TextViewer

##############################################################################
# Textual fspicker imports.
from textual_fspicker import FileSave

##############################################################################
# Local imports.
from ...peps import API
from ..data import PEP, cache_dir


##############################################################################
class PEPViewer(ModalScreen[None]):
    """A modal screen for viewing a PEP's source."""

    CSS = """
    PEPViewer {
        align: center middle;

        &> Vertical {
            width: 80%;
            height: 80%;
            max-height: 80%;
            background: $panel;
            border: solid $border;
        }

        TextViewer {
            color: $text-muted;
            height: 1fr;
            scrollbar-background: $panel;
            scrollbar-background-hover: $panel;
            scrollbar-background-active: $panel;
            &:focus {
                color: $text;
            }
        }

        #buttons {
            height: auto;
            align-horizontal: right;
            border-top: solid $border;
        }

        Button {
            margin-right: 1;
        }
    }
    """

    BINDINGS = [
        ("ctrl+c", "copy"),
        ("ctrl+r", "refresh"),
        ("ctrl+s", "save"),
        ("escape", "close"),
    ]

    def __init__(self, pep: PEP) -> None:
        """Initialise the dialog.

        Args:
            pep: The PEP to view.
        """
        super().__init__()
        self._pep = pep
        """The PEP to view."""

    def compose(self) -> ComposeResult:
        """Compose the dialog's content."""
        with Vertical() as dialog:
            dialog.border_title = f"PEP{self._pep.number}"
            yield TextViewer()
            with Horizontal(id="buttons"):
                yield Button(add_key("Copy", "^c", self), id="copy")
                yield Button(add_key("Save", "^s", self), id="save")
                yield Button(add_key("Refresh", "^r", self), id="refresh")
                yield Button(add_key("Close", "Esc", self), id="close")

    @property
    def _cache_name(self) -> Path:
        """The name of the file that is the cached version of the PEP source."""
        return cache_dir() / API.pep_file(self._pep.number)

    @work
    async def _download_text(self) -> None:
        """Download the text of the PEP.

        Notes:
            Once downloaded a local copy will be saved. Subsequently, when
            attempting to download the PEP, this local copy will be used
            instead.
        """
        (text := self.query_one(TextViewer)).loading = True
        pep_source = ""

        if self._cache_name.exists():
            try:
                pep_source = self._cache_name.read_text(encoding="utf-8")
            except IOError:
                pass

        if not pep_source:
            try:
                self._cache_name.write_text(
                    pep_source := await API().get_pep(self._pep.number),
                    encoding="utf-8",
                )
            except IOError:
                pass
            except API.RequestError as error:
                pep_source = "Error downloading PEP source"
                self.notify(
                    str(error), title="Error downloading PEP source", severity="error"
                )

        text.text = pep_source
        text.loading = False
        self.set_focus(text)

    def on_mount(self) -> None:
        """Populate the dialog once the DOM is ready."""
        self._download_text()

    @on(Button.Pressed, "#close")
    def action_close(self) -> None:
        """Close the dialog."""
        self.dismiss(None)

    @on(Button.Pressed, "#refresh")
    def action_refresh(self) -> None:
        """Refresh the PEP source."""
        try:
            self._cache_name.unlink(missing_ok=True)
        except IOError:
            pass
        self._download_text()

    @on(Button.Pressed, "#copy")
    async def action_copy(self) -> None:
        """Copy PEP text to the clipboard."""
        await self.query_one(TextArea).run_action("copy")

    @on(Button.Pressed, "#save")
    @work
    async def action_save(self) -> None:
        """Save the source of the PEP to a file."""
        if target := await self.app.push_screen_wait(
            FileSave(
                default_file=API.pep_file(self._pep.number),
                cancel_button=partial(add_key, key="Esc", context=self),
            )
        ):
            if target.exists() and not await self.app.push_screen_wait(
                Confirm(
                    "Overwrite?", f"{target}\n\nAre you sure you want to overwrite?"
                )
            ):
                return
            try:
                target.write_text(self.query_one(TextArea).text, encoding="utf-8")
            except IOError as error:
                self.notify(str(error), title="Save Failed", severity="error")
                return
            self.notify(str(target), title="Saved")


### pep_viewer.py ends here



"""A dialog for editing a PEP's notes."""

##############################################################################
# Textual imports.
from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, TextArea

##############################################################################
# Textual enhanced imports.
from textual_enhanced.tools import add_key

##############################################################################
# Local imports.
from ..data import PEP


##############################################################################
class NotesEditor(ModalScreen[str | None]):
    """A modal screen for editing some notes."""

    CSS = """
    NotesEditor {
        align: center middle;

        &> Vertical {
            width: 60%;
            max-width: 80;
            height: auto;
            background: $panel;
            border: solid $border;
        }

        TextArea, TextArea:focus {
            background: transparent;
            height: 20;
            padding: 0;
            border: none;
            scrollbar-background: $panel;
            scrollbar-background-hover: $panel;
            scrollbar-background-active: $panel;
            & > .text-area--cursor-line {
               background: transparent;
            }
        }

        #buttons {
            height: auto;
            align-horizontal: right;
            border-top: solid $border;
        }

        Button {
            margin-right: 1;
        }
    }
    """

    BINDINGS = [("escape", "cancel"), ("f2", "save")]

    def __init__(self, pep: PEP) -> None:
        """Initialise the dialog.

        Args:
            pep: The PEP to edit the notes for.
        """
        super().__init__()
        self._pep = pep
        """The PEP whose notes are being edited."""

    def compose(self) -> ComposeResult:
        """Compose the dialog's content."""
        with Vertical() as dialog:
            dialog.border_title = f"Notes for PEP{self._pep.number}"
            yield TextArea(self._pep.notes)
            with Horizontal(id="buttons"):
                yield Button(add_key("Save", "F2", self), id="save")
                yield Button(add_key("Cancel", "Esc", self), id="cancel")

    @on(Button.Pressed, "#save")
    def action_save(self) -> None:
        """Save the notes."""
        self.dismiss(self.query_one(TextArea).text)

    @on(Button.Pressed, "#cancel")
    def action_cancel(self) -> None:
        """Cancel the edit of the notes."""
        self.dismiss(None)


### notes_editor.py ends here



"""Code relating to the application's configuration file."""

##############################################################################
# Python imports.
from contextlib import contextmanager
from dataclasses import asdict, dataclass, field
from functools import lru_cache
from json import dumps, loads
from pathlib import Path
from typing import Iterator

##############################################################################
# Local imports.
from .locations import config_dir
from .peps import SortOrder


##############################################################################
@dataclass
class Configuration:
    """The configuration data for the application."""

    theme: str | None = None
    """The theme for the application."""

    sort_types_by_count: bool = True
    """Sort the navigation panel types by their count?"""

    sort_statuses_by_count: bool = True
    """Sort the navigation panel statuses by their count?"""

    sort_python_versions_by_count: bool = True
    """Sort the navigation panel Python versions by their count?"""

    sort_authors_by_count: bool = True
    """Sort the navigation panel authors by their count?"""

    details_visble: bool = False
    """Is the PEP details panel visible?"""

    peps_sort_order: SortOrder = "number"
    """The sort order of PEPs."""

    peps_sort_reversed: bool = False
    """Should the PEPs sort order be reversed?"""

    bindings: dict[str, str] = field(default_factory=dict)
    """Command keyboard binding overrides."""


##############################################################################
def configuration_file() -> Path:
    """The path to the file that holds the application configuration.

    Returns:
        The path to the configuration file.
    """
    return config_dir() / "configuration.json"


##############################################################################
def save_configuration(configuration: Configuration) -> Configuration:
    """Save the given configuration.

    Args:
        The configuration to store.

    Returns:
        The configuration.
    """
    load_configuration.cache_clear()
    configuration_file().write_text(
        dumps(asdict(configuration), indent=4), encoding="utf-8"
    )
    return load_configuration()


##############################################################################
@lru_cache(maxsize=None)
def load_configuration() -> Configuration:
    """Load the configuration.

    Returns:
        The configuration.

    Note:
        As a side-effect, if the configuration doesn't exist a default one
        will be saved to storage.

        This function is designed so that it's safe and low-cost to
        repeatedly call it. The configuration is cached and will only be
        loaded from storage when necessary.
    """
    source = configuration_file()
    return (
        Configuration(**loads(source.read_text(encoding="utf-8")))
        if source.exists()
        else save_configuration(Configuration())
    )


##############################################################################
@contextmanager
def update_configuration() -> Iterator[Configuration]:
    """Context manager for updating the configuration.

    Loads the configuration and makes it available, then ensures it is
    saved.

    Example:
        ```python
        with update_configuration() as config:
            config.meaning = 42
        ```
    """
    configuration = load_configuration()
    try:
        yield configuration
    finally:
        save_configuration(configuration)


### config.py ends here



"""Provides the main screen for the application."""

##############################################################################
# Python imports.
from argparse import Namespace
from dataclasses import dataclass
from json import dumps, loads
from webbrowser import open as visit_url

##############################################################################
# Textual imports.
from textual import on, work
from textual.app import ComposeResult
from textual.message import Message
from textual.reactive import var
from textual.widgets import Footer, Header

##############################################################################
# Textual enhanced imports.
from textual_enhanced.commands import Command, Help, Quit
from textual_enhanced.dialogs import HelpScreen, ModalInput
from textual_enhanced.screen import EnhancedScreen

##############################################################################
# Local imports.
from ... import __version__
from ...peps import API
from ..commands import (
    ChangeTheme,
    EditNotes,
    Escape,
    FindPEP,
    RedownloadPEPs,
    Search,
    SearchAuthor,
    SearchPythonVersion,
    SearchStatus,
    SearchType,
    ShowAll,
    SortByCreated,
    SortByNumber,
    SortByTitle,
    ToggleAuthorsSortOrder,
    TogglePEPDetails,
    TogglePythonVersionsSortOrder,
    ToggleSortOrder,
    ToggleStatusesSortOrder,
    ToggleTypesSortOrder,
    ViewPEP,
)
from ..data import (
    PEP,
    Containing,
    Notes,
    PEPs,
    WithAuthor,
    WithPythonVersion,
    WithStatus,
    WithType,
    load_configuration,
    pep_data,
    update_configuration,
)
from ..messages import (
    GotoPEP,
    ShowAuthor,
    ShowPythonVersion,
    ShowStatus,
    ShowType,
    VisitPEP,
)
from ..providers import (
    AuthorCommands,
    MainCommands,
    PEPsCommands,
    PythonVersionCommands,
    StatusCommands,
    TypeCommands,
)
from ..widgets import Navigation, PEPDetails, PEPsView
from .notes_editor import NotesEditor
from .pep_viewer import PEPViewer


##############################################################################
class Main(EnhancedScreen[None]):
    """The main screen for the application."""

    TITLE = f"Peplum v{__version__}"

    HELP = """
    ## Main application keys and commands

    The following keys and commands can be used anywhere here on the main screen.
    """

    DEFAULT_CSS = """
    Main {
        layout: horizontal;

        .panel {
            height: 1fr;
            border: none;
            border-left: round $border 50%;
            background: $surface;
            padding-right: 0;
            scrollbar-background: $surface;
            scrollbar-background-hover: $surface;
            scrollbar-background-active: $surface;
            &:focus, &:focus-within {
                border: none;
                border-left: round $border;
                background: $panel 80%;
                scrollbar-background: $panel;
                scrollbar-background-hover: $panel;
                scrollbar-background-active: $panel;
            }
            & > .option-list--option {
                padding: 0 1;
            }
        }

        Navigation {
            width: 2fr;
        }

        PEPsView {
            width: 8fr;
            scrollbar-gutter: stable;
        }

        PEPDetails {
            width: 3fr;
            display: none;
        }

        &.details-visible {
            PEPsView {
                width: 5fr;
            }
            PEPDetails {
                display: block;
            }
        }
    }
    """

    COMMAND_MESSAGES = (
        # Keep these together as they're bound to function keys and destined
        # for the footer.
        Help,
        EditNotes,
        TogglePEPDetails,
        ViewPEP,
        Quit,
        RedownloadPEPs,
        # Everything else.
        ChangeTheme,
        Escape,
        FindPEP,
        Search,
        SearchAuthor,
        SearchPythonVersion,
        SearchStatus,
        SearchType,
        ShowAll,
        SortByCreated,
        SortByNumber,
        SortByTitle,
        ToggleAuthorsSortOrder,
        TogglePythonVersionsSortOrder,
        ToggleSortOrder,
        ToggleStatusesSortOrder,
        ToggleTypesSortOrder,
    )

    BINDINGS = Command.bindings(*COMMAND_MESSAGES)

    COMMANDS = {MainCommands}

    all_peps: var[PEPs] = var(PEPs)
    """All the PEPs that we know about."""

    active_peps: var[PEPs] = var(PEPs)
    """The currently-active set of PEPs."""

    selected_pep: var[PEP | None] = var(None)
    """The currently-selected PEP."""

    notes: var[Notes] = var(Notes)
    """The user's notes about PEPs."""

    def __init__(self, arguments: Namespace) -> None:
        """Initialise the main screen.

        Args:
            arguments: The arguments passed to the application on the command line.
        """
        self._arguments = arguments
        """The arguments passed on the command line."""
        super().__init__()
        self._jump_to_on_load: str | None = self._arguments.pep
        """A PEP to jump to once the display is loaded."""

    def compose(self) -> ComposeResult:
        """Compose the content of the main screen."""
        yield Header()
        yield Navigation(load_configuration(), classes="panel").data_bind(
            Main.all_peps, Main.active_peps
        )
        yield PEPsView(classes="panel").data_bind(Main.active_peps)
        yield PEPDetails(classes="panel").data_bind(pep=Main.selected_pep)
        yield Footer()

    @dataclass
    class Loaded(Message):
        """A message sent when PEP data is loaded."""

        peps: PEPs
        """The PEP data that was loaded."""

    @work(thread=True)
    def load_pep_data(self) -> None:
        """Load the local copy of the PEP data."""
        if not pep_data().exists():
            return
        try:
            self.notes.load()
            self.post_message(
                self.Loaded(
                    PEPs(
                        PEP.from_storage(pep, self.notes)
                        for pep in loads(pep_data().read_text()).values()
                    )
                )
            )
        except IOError as error:
            self.notify(str(error), title="Error loading PEP data", severity="error")

    @work(thread=True)
    async def download_pep_data(self) -> None:
        """Download a fresh copy of the PEP data."""
        # Get the raw data from the API.
        try:
            raw_data = await API().get_peps()
        except API.Error as error:
            self.notify(str(error), title="API Error", severity="error", timeout=8)
            return
        # Store the raw data.
        try:
            pep_data().write_text(dumps(raw_data, indent=4), encoding="utf-8")
        except IOError as error:
            self.notify(str(error), title="Error saving PEP data", severity="error")
            return
        # Now kick off loading the raw data.
        self.notify("Fresh PEP data downloaded from the PEP API")
        self.load_pep_data()

    @staticmethod
    def _extract_pep(pep: str) -> int | None:
        """Try and extract a PEP number from a string.

        Args:
            pep: A string that should contain a PEP number.

        Returns:
            A PEP number or [`None`][None] if one could not be found.

        Notes:
            The likes of `2342` and `PEP2342` are handled.
        """
        try:
            return int(pep.strip().upper().removeprefix("PEP"))
        except ValueError:
            return None

    @on(Loaded)
    def load_fresh_peps(self, message: Loaded) -> None:
        """React to a fresh set of PEPs being made available.

        Args:
            message: The message letting us know we have fresh PEPs.
        """
        if len(message.peps.authors) == 0:
            self.notify(
                "You likely have a cached copy of the older version of the PEP data; a redownload is recommended.",
                severity="warning",
                timeout=8,
            )
        config = load_configuration()
        self.all_peps = message.peps.sorted_by(config.peps_sort_order).reversed(
            config.peps_sort_reversed
        )
        if self._jump_to_on_load is not None:
            if (pep := self._extract_pep(self._jump_to_on_load)) is not None:
                self.post_message(GotoPEP(pep))
            self._jump_to_on_load = None

    def on_mount(self) -> None:
        """Configure the application once the DOM is mounted."""
        # The caller has passed sorting preferences on the command line;
        # let's get them into the configuration before anything else kicks
        # off.
        if self._arguments.sort_by is not None:
            with update_configuration() as config:
                config.peps_sort_reversed = self._arguments.sort_by[0] == "~"
                config.peps_sort_order = self._arguments.sort_by.removeprefix("~")
        self.set_class(load_configuration().details_visble, "details-visible")
        # On startup, if we've got local PEP data...
        if pep_data().exists():
            # ...load and display that.
            self.load_pep_data()
        else:
            # Given we've got no local data at all, let's force an attempt
            # to download from the API.
            self.download_pep_data()

    def watch_all_peps(self) -> None:
        """React to the full set of PEPs being updated."""
        self.active_peps = self.all_peps
        PEPsCommands.peps = self.all_peps

    def watch_active_peps(self) -> None:
        """React to the active PEPs being updated."""
        self.sub_title = f"{self.active_peps.description} ({len(self.active_peps)})"
        AuthorCommands.active_peps = self.active_peps
        PythonVersionCommands.active_peps = self.active_peps
        StatusCommands.active_peps = self.active_peps
        TypeCommands.active_peps = self.active_peps

    @on(PEPsView.PEPHighlighted)
    def select_pep(self, message: PEPsView.PEPHighlighted) -> None:
        """Make the currently-selected PEP the one to view."""
        self.selected_pep = message.pep

    @on(PEPsView.Empty)
    def deselect_pep(self) -> None:
        """Handle there being no highlighted PEP."""
        self.selected_pep = None

    @on(ShowAll)
    def action_show_all_command(self) -> None:
        """Show all PEPs."""
        self.active_peps = self.all_peps

    @on(ShowType)
    def show_type(self, command: ShowType) -> None:
        """Filter the PEPs by a given type.

        Args:
            command: The command requesting the filter.
        """
        self.active_peps &= WithType(command.type)

    @on(ShowStatus)
    def show_status(self, command: ShowStatus) -> None:
        """Filter the PEPs by a given status.

        Args:
            command: The command requesting the filter.
        """
        self.active_peps &= WithStatus(command.status)

    @on(ShowPythonVersion)
    def show_python_version(self, command: ShowPythonVersion) -> None:
        """Filter the PEPs by a given Python version.

        Args:
            command: The command requesting the filter.
        """
        self.active_peps &= WithPythonVersion(command.version)

    @on(ShowAuthor)
    def show_author(self, command: ShowAuthor) -> None:
        """Filter the PEPs by a given author.

        Args:
            command: The command requesting the filter.
        """
        self.active_peps &= WithAuthor(command.author)

    @on(GotoPEP)
    def goto_pep(self, command: GotoPEP) -> None:
        """Visit a specific PEP by its number.

        Args:
            command: The command requesting the PEP by number.
        """
        if command.number in self.active_peps:
            self.query_one(PEPsView).goto_pep(command.number)
        elif command.number in self.all_peps:
            self.notify(
                f"PEP{command.number} wasn't in the active filter; switching to all PEPs...",
                severity="warning",
            )
            self.active_peps = self.all_peps
            self.call_after_refresh(self.query_one(PEPsView).goto_pep, command.number)
        else:
            self.notify(
                f"PEP{command.number} doesn't exist. Perhaps you'll be the one to write it?",
                title="No such PEP",
                severity="error",
            )

    @on(VisitPEP)
    def visit_pep(self, command: VisitPEP) -> None:
        """Visit a given PEP's webpage.

        Args:
            command: The command requesting the visit.
        """
        if command.pep.url:
            visit_url(command.pep.url)
        else:
            self.notify(f"PEP{command.pep.number} has no associated URL")

    @on(FindPEP)
    def action_find_pep_command(self) -> None:
        """Find a PEP and jump to it."""
        self.show_palette(PEPsCommands)

    @on(Search)
    @work
    async def action_search_command(self) -> None:
        """Free-text search within the PEPs."""
        if search_text := await self.app.push_screen_wait(
            ModalInput("Case-insensitive text to look for in the PEPs")
        ):
            self.active_peps = self.active_peps & Containing(search_text)

    @on(SearchAuthor)
    def action_search_author_command(self) -> None:
        """Search for an author and use them as a filter."""
        self.show_palette(AuthorCommands)

    @on(SearchPythonVersion)
    def action_search_python_version_command(self) -> None:
        """Search for a Python version and then use it as a filter."""
        self.show_palette(PythonVersionCommands)

    @on(SearchStatus)
    def action_search_status_command(self) -> None:
        """Search for a status and use it as a filter."""
        self.show_palette(StatusCommands)

    @on(SearchType)
    def action_search_type_command(self) -> None:
        """Search for a PEP type and then use it as a filter."""
        self.show_palette(TypeCommands)

    @on(RedownloadPEPs)
    def action_redownload_peps_command(self) -> None:
        """Redownload PEPs from the API."""
        self.download_pep_data()

    @on(Help)
    def action_help_command(self) -> None:
        """Toggle the display of the help panel."""
        self.app.push_screen(HelpScreen(self))

    @on(ChangeTheme)
    def action_change_theme_command(self) -> None:
        """Show the theme picker."""
        self.app.search_themes()

    @on(Quit)
    def action_quit_command(self) -> None:
        """Quit the application."""
        self.app.exit()

    @on(Escape)
    def action_escape_command(self) -> None:
        """Handle escaping.

        The action's approach is to step-by-step back out from the 'deepest'
        level to the topmost, and if we're at the topmost then exit the
        application.
        """
        if self.focused == self.query_one(PEPsView):
            self.set_focus(self.query_one(Navigation))
        elif (
            self.focused
            and self.query_one(PEPDetails) in self.focused.ancestors_with_self
        ):
            self.set_focus(self.query_one(PEPsView))
        else:
            self.app.exit()

    @on(TogglePEPDetails)
    def action_toggle_pep_details_command(self) -> None:
        """Toggle the display of the PEP details panel."""
        self.toggle_class("details-visible")
        with update_configuration() as config:
            config.details_visble = self.has_class("details-visible")

    @on(ToggleTypesSortOrder)
    def action_toggle_types_sort_order_command(self) -> None:
        """Toggle the sort order of the types."""
        with update_configuration() as config:
            config.sort_types_by_count = not config.sort_types_by_count
            self.query_one(Navigation).sort_types_by_count = config.sort_types_by_count

    @on(ToggleStatusesSortOrder)
    def action_toggle_statuses_sort_order_command(self) -> None:
        """Toggle the sort order of the statuses."""
        with update_configuration() as config:
            config.sort_statuses_by_count = not config.sort_statuses_by_count
            self.query_one(
                Navigation
            ).sort_statuses_by_count = config.sort_statuses_by_count

    @on(TogglePythonVersionsSortOrder)
    def action_toggle_python_versions_sort_order_command(self) -> None:
        """Toggle the sort order of the Python Versions."""
        with update_configuration() as config:
            config.sort_python_versions_by_count = (
                not config.sort_python_versions_by_count
            )
            self.query_one(
                Navigation
            ).sort_python_versions_by_count = config.sort_python_versions_by_count

    @on(ToggleAuthorsSortOrder)
    def action_toggle_authors_sort_order_command(self) -> None:
        """Toggle the sort order of the authors."""
        with update_configuration() as config:
            config.sort_authors_by_count = not config.sort_authors_by_count
            self.query_one(
                Navigation
            ).sort_authors_by_count = config.sort_authors_by_count

    @on(SortByCreated)
    def action_sort_by_created_command(self) -> None:
        """Sort the PEPs by their date created."""
        with update_configuration() as config:
            config.peps_sort_order = "created"
            self.active_peps = self.active_peps.sorted_by(config.peps_sort_order)

    @on(SortByNumber)
    def action_sort_by_number_command(self) -> None:
        """Sort the PEPs by their number."""
        with update_configuration() as config:
            config.peps_sort_order = "number"
            self.active_peps = self.active_peps.sorted_by(config.peps_sort_order)

    @on(SortByTitle)
    def action_sort_by_title_command(self) -> None:
        """Sort the PEPs by their title."""
        with update_configuration() as config:
            config.peps_sort_order = "title"
            self.active_peps = self.active_peps.sorted_by(config.peps_sort_order)

    @on(ToggleSortOrder)
    def action_toggle_sort_order_command(self) -> None:
        """Toggle the current sort order direction of the PEPs."""
        with update_configuration() as config:
            config.peps_sort_reversed = not config.peps_sort_reversed
            self.active_peps = self.active_peps.reversed(config.peps_sort_reversed)

    @work(thread=True)
    def _save_notes(self) -> None:
        """Save the notes."""
        try:
            self.notes.save()
        except IOError as error:
            self.notify(
                str(error), title="Unable to save notes", severity="error", timeout=8
            )

    @on(EditNotes)
    @work
    async def action_edit_notes_command(self) -> None:
        """Edit the notes for the currently-highlighted PEP."""
        if self.selected_pep is None:
            self.notify("Highlight a PEP to edit its notes.", severity="warning")
            return
        if (
            notes := await self.app.push_screen_wait(NotesEditor(self.selected_pep))
        ) is not None:
            self.notes[self.selected_pep.number] = notes
            self._save_notes()
            self.active_peps = self.active_peps.rebuild_from(
                self.all_peps.patch_pep(self.selected_pep.annotate(notes=notes))
            )

    @on(ViewPEP)
    def action_view_pep_command(self) -> None:
        """View the currently-highlighted PEP's source."""
        if self.selected_pep is None:
            self.notify("Highlight a PEP to view it.", severity="warning")
            return
        if self.selected_pep.number == 0:
            self.notify("PEP0 has no source to view.", severity="warning")
            return
        self.app.push_screen(PEPViewer(self.selected_pep))


### main.py ends here



"""Provides a class for letting the user view a list of PEPs."""

##############################################################################
# Python imports.
from dataclasses import dataclass
from typing import Final

##############################################################################
# Rich imports.
from rich.console import Group
from rich.markup import escape
from rich.rule import Rule
from rich.table import Table

##############################################################################
# Textual imports.
from textual import on
from textual.message import Message
from textual.reactive import var
from textual.widgets.option_list import Option

##############################################################################
# Textual enhanced imports.
from textual_enhanced.widgets import EnhancedOptionList

##############################################################################
# Local imports.
from ..data import PEP, PEPs
from ..messages import VisitPEP


##############################################################################
class PEPView(Option):
    """Option for viewing a single PEP."""

    RULE: Final[Rule] = Rule(style="dim")
    """The rule to place at the end of each view."""

    def __init__(self, pep: PEP) -> None:
        """Initialise the object.

        Args:
            pep: The PEP to view.
        """

        title = Table.grid(expand=True)
        title.add_column(width=6)
        title.add_column(justify="left", ratio=2)
        title.add_column(justify="right", width=28)
        title.add_row(
            f"[bold]{pep.number}[/]",
            escape(pep.title),
            f"[dim]{pep.type}, {pep.status}[/]",
        )

        info = Table.grid(expand=True)
        info.add_column(width=6)
        info.add_column(ratio=1)
        info.add_column(width=11, justify="right")
        info.add_row(
            "", f"[dim]{', '.join(pep.author_names)}[/]", f"[dim]{pep.created}[/]"
        )

        self._pep = pep
        """The PEP that this option is showing."""

        super().__init__(Group(title, info, self.RULE), id=self.make_id(pep.number))

    @staticmethod
    def make_id(number: int) -> str:
        """Make an ID from a given PEP number.

        Args:
            number: The number of the PEP to make an ID for.

        Returns:
            The ID for the PEP.
        """
        return f"PEP{number}"

    @property
    def pep(self) -> PEP:
        """The PEP associated with this option."""
        return self._pep


##############################################################################
class PEPsView(EnhancedOptionList):
    """A widget for viewing a list of PEPs."""

    HELP = """
    ## The PEPs List

    This is a list of all PEPs that match your current filter.
    """

    active_peps: var[PEPs] = var(PEPs)
    """The currently-active collection of PEPs."""

    class Empty(Message):
        """A message sent when the PEPs view falls empty."""

    def watch_active_peps(self) -> None:
        """React to the PEPs being changed."""
        with self.preserved_highlight:
            self.clear_options().add_options(PEPView(pep) for pep in self.active_peps)
        if not self.option_count:
            self.post_message(self.Empty())

    @dataclass
    class PEPHighlighted(Message):
        """A message that is posted when a PEP is highlighted by the user."""

        pep: PEP
        """The highlighted PEP."""

    @on(EnhancedOptionList.OptionHighlighted)
    def select_pep(self, message: EnhancedOptionList.OptionSelected) -> None:
        """Send a message to say a particular PEP has been selected."""
        message.stop()
        assert isinstance(message.option, PEPView)
        self.post_message(self.PEPHighlighted(message.option.pep))

    @on(EnhancedOptionList.OptionSelected)
    def visit_pep(self, message: EnhancedOptionList.OptionSelected) -> None:
        """Send a message to say the user wants to visit a PEP's webpage."""
        message.stop()
        assert isinstance(message.option, PEPView)
        self.post_message(VisitPEP(message.option.pep))

    def goto_pep(self, pep: int) -> None:
        """Jump to a specific PEP."""
        self.highlighted = self.get_option_index(PEPView.make_id(pep))
        self.screen.set_focus(self)


### peps_view.py ends here



"""Commands for affecting navigation sort ordering."""

##############################################################################
# Textual enhanced imports.
from textual_enhanced.commands import Command


##############################################################################
class ToggleTypesSortOrder(Command):
    """Toggle the sort order of types in the navigation panel"""

    BINDING_KEY = "T"


##############################################################################
class ToggleStatusesSortOrder(Command):
    """Toggle the sort order of the statuses in the navigation panel"""

    BINDING_KEY = "S"


##############################################################################
class TogglePythonVersionsSortOrder(Command):
    """Toggle the sort order of Python versions in the navigation panel"""

    BINDING_KEY = "V"


##############################################################################
class ToggleAuthorsSortOrder(Command):
    """Toggle the sort order of the authors in the navigation panel"""

    BINDING_KEY = "A"


### navigation_sorting.py ends here



"""The main navigation panel."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from typing import Callable

##############################################################################
# Rich imports.
from rich.console import Group, RenderableType
from rich.rule import Rule
from rich.table import Table

##############################################################################
# Textual imports.
from textual import on
from textual.message import Message
from textual.reactive import var
from textual.widgets import OptionList
from textual.widgets.option_list import Option

##############################################################################
# Textual enhanced imports.
from textual_enhanced.widgets import EnhancedOptionList

##############################################################################
# Typing exception imports.
from typing_extensions import Self

##############################################################################
# Local imports.
from ..commands import ShowAll
from ..data import (
    AuthorCount,
    Configuration,
    PEPCount,
    PEPs,
    PythonVersionCount,
    StatusCount,
    TypeCount,
)
from ..messages import ShowAuthor, ShowPythonVersion, ShowStatus, ShowType


##############################################################################
class Title(Option):
    """Option for showing a title."""

    def __init__(self, title: str) -> None:
        """Initialise the object.

        Args:
            title: The title to show.
        """
        super().__init__(
            Group("", Rule(title, style="bold dim")),
            disabled=True,
            id=f"_title_{title}",
        )


##############################################################################
class CountView(Option):
    """Base class for options that show a count."""

    def count_prompt(self, caption: str, count: int) -> RenderableType:
        """Create a prompt.

        Args:
            caption: The caption for the prompt.
            count: The count for the prompt.

        Returns:
            The prompt.
        """
        prompt = Table.grid(expand=True)
        prompt.add_column(ratio=1)
        prompt.add_column(justify="right")
        prompt.add_row(caption, f"[dim i]{count}[/]")
        return prompt

    @property
    def command(self) -> Message:
        """The command to send when this option is selected."""
        raise NotImplementedError


##############################################################################
class AllView(CountView):
    """Option used to signify that we should view all PEPs."""

    def __init__(self, peps: PEPs, key: str, key_colour: str | None) -> None:
        """Initialise the object.

        Args:
            peps: The full collection of PEPs.
        """
        super().__init__(
            self.count_prompt(f"All [{(key_colour or 'dim')}]\\[{key}][/]", len(peps)),
            id=f"_all_peps",
        )

    @property
    def command(self) -> Message:
        """The command to send when this option is selected."""
        return ShowAll()


##############################################################################
class TypeView(CountView):
    """Option for showing a PEP type."""

    def __init__(self, pep_type: TypeCount) -> None:
        """Initialise the object.

        Args:
            pep_type: The details of the PEP type to show.
        """
        self._type = pep_type
        """The details of the type to show."""
        super().__init__(
            self.count_prompt(pep_type.type, pep_type.count),
            id=f"_type_{pep_type.type}",
        )

    @property
    def command(self) -> Message:
        """The command to send when this option is selected."""
        return ShowType(self._type.type)


##############################################################################
class StatusView(CountView):
    """Option for showing a PEP status."""

    def __init__(self, status: StatusCount) -> None:
        """Initialise the object.

        Args:
            status: The details of the PEP status to show.
        """
        self._status = status
        """The details of the status to show."""
        super().__init__(
            self.count_prompt(status.status, status.count),
            id=f"_status_{status.status}",
        )

    @property
    def command(self) -> Message:
        """The command to send when this option is selected."""
        return ShowStatus(self._status.status)


##############################################################################
class PythonVersionView(CountView):
    """Option for showing a Python version."""

    def __init__(self, version: PythonVersionCount) -> None:
        """Initialise the object.

        Args:
            version: The details of the PEP Python version to show.
        """
        self._version = version
        """The Python version to show."""
        super().__init__(
            self.count_prompt(version.version or f"[dim i]None[/]", version.count),
            id=f"_python_version_{version.version}",
        )

    @property
    def command(self) -> Message:
        """The command to send when this option is selected."""
        return ShowPythonVersion(self._version.version)


##############################################################################
class AuthorView(CountView):
    """Option for showing a PEP author."""

    def __init__(self, author: AuthorCount) -> None:
        """Initialise the object.

        Args:
            author: The details of the PEP author to show.
        """
        self._author = author
        """The details of the author to show."""
        super().__init__(
            self.count_prompt(author.author, author.count),
            id=f"_author_{author.author}",
        )

    @property
    def command(self) -> Message:
        """The command to send when this option is selected."""
        return ShowAuthor(self._author.author)


##############################################################################
class Navigation(EnhancedOptionList):
    """The main navigation panel."""

    HELP = """
    ## Navigation Panel

    Select items in this panel to filter the list of PEPs.
    """

    all_peps: var[PEPs] = var(PEPs)
    """The collection of all known PEPs."""

    active_peps: var[PEPs] = var(PEPs)
    """The currently-active collection of PEPs."""

    sort_types_by_count: var[bool] = var(True)
    """Sort the types by their count?"""

    sort_statuses_by_count: var[bool] = var(True)
    """Sort the statuses by their count?"""

    sort_python_versions_by_count: var[bool] = var(True)
    """Sort the Python versions by their count?"""

    sort_authors_by_count: var[bool] = var(True)
    """Sort the authors by their count?"""

    def __init__(
        self, config: Configuration, id: str | None = None, classes: str | None = None
    ):
        """Initialise the widget.

        Args:
            config: THe configuration for the application.
            id: The ID for the widget.
            classes: The classes for the widget.
        """
        super().__init__(id=id, classes=classes)
        self.set_reactive(Navigation.sort_types_by_count, config.sort_types_by_count)
        self.set_reactive(
            Navigation.sort_statuses_by_count, config.sort_statuses_by_count
        )
        self.set_reactive(
            Navigation.sort_python_versions_by_count,
            config.sort_python_versions_by_count,
        )
        self.set_reactive(
            Navigation.sort_authors_by_count, config.sort_authors_by_count
        )

    def on_mount(self) -> None:
        """Configure the widget once the DOM is mounted."""
        self.app.theme_changed_signal.subscribe(self, lambda _: self.repopulate())

    def add_main(self) -> Self:
        """Add the main navigation options.

        Returns:
            Self.
        """
        return self.add_option(
            AllView(
                self.all_peps,
                key=ShowAll.key_binding(),
                key_colour=None
                if self.app.current_theme is None
                else self.app.current_theme.accent,
            )
        )

    @staticmethod
    def _filter_key(by_count: bool) -> Callable[[PEPCount], int | PEPCount]:
        """Get a key for sorting a filter in navigation.

        Args:
            by_count: Should we sort by count?

        Returns:
            A function to get the correct key to sort on.
        """

        def _key(count: PEPCount) -> int | PEPCount:
            return -count.count if by_count else count

        return _key

    def add_types(self) -> Self:
        """Add the PEP types to navigation.

        Returns:
            Self.
        """
        if self.active_peps:
            self.add_option(Title("Type"))
            for pep_type in sorted(
                self.active_peps.types, key=self._filter_key(self.sort_types_by_count)
            ):
                self.add_option(TypeView(pep_type))
        return self

    def add_statuses(self) -> Self:
        """Add the PEP statuses to navigation.

        Returns:
            Self.
        """
        if self.active_peps:
            self.add_option(Title("Status"))
            for status in sorted(
                self.active_peps.statuses,
                key=self._filter_key(self.sort_statuses_by_count),
            ):
                self.add_option(StatusView(status))
        return self

    def add_python_versions(self) -> Self:
        """Add the PEP python versions to navigation.

        Returns:
            Self.
        """
        if self.active_peps:
            self.add_option(Title("Python Version"))
            for version in sorted(
                self.active_peps.python_versions,
                key=self._filter_key(self.sort_python_versions_by_count),
            ):
                self.add_option(PythonVersionView(version))
        return self

    def add_authors(self) -> Self:
        """Add the PEP authors to navigation.

        Returns:
            Self.
        """
        if self.active_peps:
            self.add_option(Title("Author"))
            for author in sorted(
                self.active_peps.authors,
                key=self._filter_key(self.sort_authors_by_count),
            ):
                self.add_option(AuthorView(author))
        return self

    def repopulate(self) -> None:
        """Repopulate navigation panel."""
        with self.preserved_highlight:
            self.clear_options().add_main().add_types().add_statuses().add_python_versions().add_authors()

    def watch_all_peps(self) -> None:
        """React to the full list of PEPs being changed."""
        self.repopulate()

    def watch_active_peps(self) -> None:
        """React to the active PEPs being changed."""
        self.repopulate()

    def watch_sort_types_by_count(self) -> None:
        """React to the types sort order being changed."""
        self.repopulate()

    def watch_sort_statuses_by_count(self) -> None:
        """React to the statuses sort order being changed."""
        self.repopulate()

    def watch_sort_python_versions_by_count(self) -> None:
        """React to the Python versions sort order being changed."""
        self.repopulate()

    def watch_sort_authors_by_count(self) -> None:
        """React to the authors sort order being changed."""
        self.repopulate()

    @on(OptionList.OptionSelected)
    def navigate(self, event: OptionList.OptionSelected) -> None:
        event.stop()
        assert isinstance(event.option, CountView)
        self.post_message(event.option.command)


### navigation.py ends here



"""Python version filtering commands for the command palette."""

##############################################################################
# Textual enhanced imports.
from textual_enhanced.commands import CommandHit, CommandHits, CommandsProvider

##############################################################################
# Local imports.
from ..data import PEPs
from ..messages import ShowPythonVersion


##############################################################################
class PythonVersionCommands(CommandsProvider):
    """A command palette provider related to Python versions."""

    active_peps: PEPs | None = None
    """The currently-active collection of PEPs to get the Python versions of."""

    @classmethod
    def prompt(cls) -> str:
        """The prompt for the command provider."""
        return (
            "Also search for PEPs related to Python version..."
            if cls.active_peps and cls.active_peps.is_filtered
            else "Search for PEPs related to Python version..."
        )

    def commands(self) -> CommandHits:
        """Provide the Python version-based command data for the command palette.

        Yields:
            The commands for the command palette.
        """
        if self.active_peps is None:
            return
        help_prefix, command_prefix = (
            ("Also filter", "Filter")
            if self.active_peps.is_filtered
            else ("Also relating to Python version", "Relating to Python version")
        )
        for version in sorted(self.active_peps.python_versions):
            if not version.version:
                yield CommandHit(
                    "Also isn't related to a specific Python version"
                    if self.active_peps.is_filtered
                    else "Isn't related to a specific Python version",
                    f"{help_prefix} to PEPs unrelated to any specific Python version (narrows down to {version.count})",
                    ShowPythonVersion(""),
                )
            else:
                yield CommandHit(
                    f"{command_prefix} {version.version}",
                    f"{help_prefix} to PEPs related to Python version {version.version} (narrows down to {version.count})",
                    ShowPythonVersion(version.version),
                )


### python_versions.py ends here



"""Provides a PEP API client class."""

##############################################################################
# Python imports.
from pathlib import Path
from ssl import SSLCertVerificationError
from typing import Any, Final

##############################################################################
# HTTPX imports.
from httpx import AsyncClient, HTTPStatusError, RequestError, Response


##############################################################################
class API:
    """API client for peps.python.org."""

    AGENT: Final[str] = "Peplum (https://github.com/davep/peplum)"
    """The agent string to use when talking to the API."""

    _URL: Final[str] = "https://peps.python.org/api/peps.json"
    """The URL of the PEP download API."""

    class Error(Exception):
        """Base class for Raindrop errors."""

    class RequestError(Error):
        """Exception raised if there was a problem making an API request."""

    def __init__(self) -> None:
        """Initialise the client object."""
        self._client_: AsyncClient | None = None
        """The internal reference to the HTTPX client."""

    @property
    def _client(self) -> AsyncClient:
        """The HTTPX client."""
        if self._client_ is None:
            self._client_ = AsyncClient()
        return self._client_

    async def _get(self, url: str) -> Response:
        """Make a GET request.

        Args:
            url: The URL to make the request of.

        Returns:
            The response.

        Raises:
            RequestError: If there was some sort of error.
        """
        try:
            response = await self._client.get(url, headers={"user-agent": self.AGENT})
        except (RequestError, SSLCertVerificationError) as error:
            raise self.RequestError(str(error)) from None

        try:
            response.raise_for_status()
        except HTTPStatusError as error:
            raise self.RequestError(str(error)) from None

        return response

    async def get_peps(self) -> dict[int, dict[str, Any]]:
        """Download a fresh list of all known PEPs.

        Returns:
            The PEP JSON data.

        Raises:
            RequestError: If there was a problem getting the PEPS.
        """
        if isinstance(
            raw_data := (
                await self._get("https://peps.python.org/api/peps.json")
            ).json(),
            dict,
        ):
            return raw_data
        raise RequestError("Unexpected data received from the PEP API")

    @staticmethod
    def pep_file(pep: int) -> Path:
        """Generate the name of the source file of a PEP.

        Args:
            pep: The number of the PEP.

        Returns:
            The name of the source file for that PEP.
        """
        return Path(f"pep-{pep:04}.rst")

    @classmethod
    def pep_url(cls, pep: int) -> str:
        """Generate the URL for the source of a PEP.

        Args:
            pep: The number of the PEP.

        Returns:
            The URL for the source of the PEP.
        """
        return f"https://raw.githubusercontent.com/python/peps/refs/heads/main/peps/{cls.pep_file(pep)}"

    async def get_pep(self, pep: int) -> str:
        """Download the text of a given PEP.

        Args:
            pep: The number of the PEP to download.

        Returns:
            The text for the PEP.
        """
        return (await self._get(self.pep_url(pep))).text


### api.py ends here



"""Test annotating a PEP."""

##############################################################################
# Pytest imports.
from pytest import fixture

##############################################################################
# Local imports.
from peplum.app.data import PEP


##############################################################################
@fixture
def example_pep() -> PEP:
    """An example PEP to work from."""
    return PEP.from_api(
        {
            "number": 1,
            "title": "PEP Purpose and Guidelines",
            "authors": "Barry Warsaw, Jeremy Hylton, David Goodger, Alyssa Coghlan",
            "discussions_to": None,
            "status": "Active",
            "type": "Process",
            "topic": "",
            "created": "13-Jun-2000",
            "python_version": None,
            "post_history": "21-Mar-2001, 29-Jul-2002, 03-May-2003, 05-May-2012, 07-Apr-2013",
            "resolution": None,
            "requires": None,
            "replaces": None,
            "superseded_by": None,
            "url": "https://peps.python.org/pep-0001/",
        }
    )


##############################################################################
def test_no_annotation(example_pep: PEP) -> None:
    """Annotating with nothing should do nothing."""
    assert id(example_pep.annotate()) == id(example_pep)


##############################################################################
def test_an_annotation(example_pep: PEP) -> None:
    """Annotating with new values result in a new instance of the PEP."""
    assert id(example_pep.annotate(notes="Changed")) != id(example_pep)


##############################################################################
def test_add_note(example_pep: PEP) -> None:
    """A note added to a PEP should be in the PEP."""
    assert example_pep.annotate(notes="test").notes == "test"


### test_pep_annotation.py ends here



"""The main commands used within the application."""

##############################################################################
# Textual enhanced imports.
from textual_enhanced.commands import Command


##############################################################################
class ChangeTheme(Command):
    """Change the application's theme"""

    BINDING_KEY = "f9"


##############################################################################
class EditNotes(Command):
    """Edit the highlighted PEP's notes"""

    BINDING_KEY = "f2"
    FOOTER_TEXT = "Notes"
    SHOW_IN_FOOTER = True


##############################################################################
class Escape(Command):
    "Back up through the panes, right to left, or exit the app if the navigation pane has focus"

    BINDING_KEY = "escape"


##############################################################################
class RedownloadPEPs(Command):
    """Redownload the list of PEPs"""

    FOOTER_TEXT = "Redownload"
    COMMAND = "Redownload All PEPs"
    BINDING_KEY = "ctrl+r"
    ACTION = "redownload_peps_command"
    SHOW_IN_FOOTER = True


##############################################################################
class TogglePEPDetails(Command):
    """Toggle the display of the PEP details panel"""

    FOOTER_TEXT = "Details"
    BINDING_KEY = "f3"
    SHOW_IN_FOOTER = True


##############################################################################
class ViewPEP(Command):
    """View the source of the currently-highlighted PEP"""

    FOOTER_TEXT = "View"
    BINDING_KEY = "f4"
    SHOW_IN_FOOTER = True


### main.py ends here



"""Provides a class for holding data about a PEP."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from dataclasses import dataclass, replace
from datetime import date
from re import Pattern, compile
from typing import Any, Final, Literal, cast

##############################################################################
# Local imports.
from .notes import Notes

##############################################################################
PEPStatus = Literal[
    "Draft",
    "Active",
    "Accepted",
    "Provisional",
    "Deferred",
    "Rejected",
    "Withdrawn",
    "Final",
    "Superseded",
]
"""The possible status values for a PEP."""

##############################################################################
PEPType = Literal[
    "Standards Track",
    "Informational",
    "Process",
]
"""The possible types of PEP."""


##############################################################################
MONTHS: Final[dict[str, int]] = {
    "jan": 1,
    "feb": 2,
    "mar": 3,
    "apr": 4,
    "may": 5,
    "jun": 6,
    "jul": 7,
    "aug": 8,
    "sep": 9,
    "oct": 10,
    "nov": 11,
    "dec": 12,
}
"""Month name to number translation table."""

##############################################################################
PEP_DATE: Final[Pattern[str]] = compile(
    r"^(?P<day>\d{2})-(?P<month>\w{3})-(?P<year>\d{4})$"
)
"""Regular expression for parsing a PEP's date."""


##############################################################################
def parse_date(date_value: str) -> date:
    """Parse the sort of date found in the PEP index.

    Args:
        date_value: The value to parse.

    Returns:
        A `date` object.

    Raises:
        ValueError: If the date string is invalid.

    Notes:
        The dates in the PEP index use a less-than-ideal date format, being
        of the form DD-MMM-YYYY where MMM is a truncated form of the month
        name in English. Because of this this function hand-parses the date
        rather than use strptime, because locales exist and flipping locale
        is problematic for various reasons.
    """
    if parsed_date := PEP_DATE.match(date_value):
        try:
            return date(
                year=int(parsed_date["year"]),
                month=MONTHS[parsed_date["month"].lower()],
                day=int(parsed_date["day"]),
            )
        except KeyError:
            raise ValueError(f"{date_value} is not a recognised date") from None
    else:
        raise ValueError(f"Can't parse {date_value} as a PEP date")


##############################################################################
DATE_ONLY: Final[Pattern[str]] = compile(r"^\d{2}-\w{3}-\d{4}$")
"""Regular expression for detecting just a date in the input."""
URL_ONLY: Final[Pattern[str]] = compile("^http")
"""Regular expression for detecting just an URL in the input."""
DATE_AND_URL: Final[Pattern[str]] = compile(
    r"^`(?P<date>\d{2}-\w{3}-\d{4}).+<(?P<url>https:.*)>`__$"
)
"""Regular expression for detecting a date and URL in the input."""


##############################################################################
@dataclass(frozen=True)
class PostHistory:
    """Details of an item in a PEP's post history."""

    date: date | None = None
    """The date of the post history."""
    url: str | None = None
    """The URL of a link to the post, if there is one."""

    @classmethod
    def from_value(cls, value: str | None) -> PostHistory | None:
        """Create a post history object from the given value.

        Args:
            value: The value to create the history from.

        Returns:
            A `PostHistory` instance, or `None` if `None` was the input.
        """
        if value is None:
            return value
        if match := DATE_ONLY.match(value):
            return PostHistory(date=parse_date(value))
        if match := URL_ONLY.match(value):
            return PostHistory(url=value)
        if match := DATE_AND_URL.match(value):
            return PostHistory(date=parse_date(match["date"]), url=match["url"])
        raise ValueError(f"Can't parse `{value}` as PostHistory")


##############################################################################
@dataclass(frozen=True)
class PEP:
    """A class that holds data about a PEP."""

    number: int
    """The number of the PEP."""
    title: str
    """The title of the PEP."""
    authors: str
    """The authors of the PEP."""
    author_names: tuple[str, ...]
    """The names of the authors of the PEP."""
    sponsor: str | None
    """The sponsor of the PEP."""
    delegate: str | None
    """The name of the PEP's delegate."""
    discussions_to: str | None
    """The location where discussions about this PEP should take place."""
    status: PEPStatus
    """The status of the PEP."""
    type: PEPType
    """The type of the PEP."""
    topic: str
    """The topic of the PEP."""
    requires: tuple[int, ...]
    """The PEPS that this PEP requires."""
    created: date
    """The date the PEP was created."""
    python_version: tuple[str, ...]
    """The Python versions this PEP relates to."""
    post_history: tuple[PostHistory, ...]
    """The dates of the posting history for the PEP."""
    resolution: PostHistory | None
    """The resolution of the PEP, if it has one."""
    replaces: tuple[int, ...]
    """The PEPs this PEP replaces."""
    superseded_by: int | None
    """The PEP that supersedes this PEP."""
    url: str
    """The URL for the PEP."""
    notes: str = ""
    """The user's notes associated with this PEP."""

    def annotate(self, *, notes: str | None = None) -> PEP:
        """Annotate the PEP.

        Args:
            notes: The optional notes to annotate the PEP with.
        """
        if notes is not None:
            return replace(self, notes=notes)
        return self

    def __contains__(self, search_text: str) -> bool:
        """Perhaps a case-insensitive search for the text anywhere in the PEP's data.

        Args:
            search_text: The text to search for.

        Returns:
            `True` if the text can be found, `False` if not.
        """
        search_text = search_text.casefold()
        return (
            search_text in str(self.number)
            or search_text in self.title.casefold()
            or search_text in " ".join(self.author_names).casefold()
            or search_text in (self.sponsor or "").casefold()
            or search_text in (self.delegate or "").casefold()
            or search_text in self.status.casefold()
            or search_text in self.type.casefold()
            or search_text in self.topic.casefold()
            or search_text in (str(pep) for pep in self.requires)
            or search_text in str(self.created)
            or search_text in " ".join(str(version) for version in self.python_version)
            or search_text in (str(pep) for pep in self.replaces)
            or search_text
            in ("" if self.superseded_by is None else str(self.superseded_by))
            or search_text in self.url.casefold()
            or search_text in self.notes.casefold()
        )

    @classmethod
    def _parse(cls, data: dict[str, Any]) -> dict[str, Any]:
        """Parse the content of a PEP from the API data.

        Args:
            data: The data from the PEP API.

        Returns:
            The data turned into locally-useful values.
        """

        def get_ints(field: str) -> tuple[int, ...]:
            if isinstance(values := data[field], str):
                return tuple(int(value) for value in values.split(","))
            return ()

        return dict(
            number=data.get("number", -1),
            title=data.get("title", ""),
            authors=data.get("authors"),
            author_names=tuple(data.get("author_names", [])),
            sponsor=data.get("sponsor"),
            delegate=data.get("delegate"),
            discussions_to=data.get("discussions_to"),
            status=cast(PEPStatus, data.get("status")),
            type=cast(PEPType, data.get("type")),
            topic=data.get("topic", ""),
            requires=get_ints("requires"),
            created=parse_date(data.get("created", "")),
            python_version=tuple(
                version.strip()
                for version in (data.get("python_version") or "").split(",")
                if version
            ),
            post_history=tuple(
                PostHistory.from_value(post.strip()) or PostHistory()
                for post in (data.get("post_history", "") or "").split(",")
                if post
            ),
            resolution=PostHistory.from_value(data.get("resolution")),
            replaces=get_ints("replaces"),
            superseded_by=None
            if data.get("superseded_by") is None
            else int(data["superseded_by"]),
            url=data.get("url", ""),
        )

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> PEP:
        """Create a PEP from the given API data.

        Args:
            data: The data to create the object from.

        Returns:
            A fresh `PEP` object created from the data.
        """
        return cls(**cls._parse(data))

    @classmethod
    def from_storage(cls, data: dict[str, Any], notes: Notes) -> PEP:
        """Create a PEP from the given data from storage.

        Args:
            data: The data to create the object from.
            notes: The local notes about PEPs.

        Returns:
            A fresh `PEP` object created from the data.
        """
        return cls(**(pep := cls._parse(data)), notes=notes[pep["number"]])


### pep.py ends here



"""Commands for locating and jumping to a PEP."""

##############################################################################
# Python imports.
from operator import attrgetter

##############################################################################
# Textual enhanced imports.
from textual_enhanced.commands import CommandHit, CommandHits, CommandsProvider

##############################################################################
# Local imports.
from ..data import PEPs
from ..messages import GotoPEP


##############################################################################
class PEPsCommands(CommandsProvider):
    """A command palette provider for finding and jumping to a PEP."""

    peps: PEPs | None = None
    """The list of PEPs to show."""

    @classmethod
    def prompt(cls) -> str:
        """The prompt for the command provider."""
        return "Jump to PEP..."

    def commands(self) -> CommandHits:
        """Provide a list of commands for jumping to a specific PEP.

        Yields:
           Commands to show in the command palette.
        """
        if self.peps is None:
            return
        for pep in sorted(self.peps, key=attrgetter("number")):
            yield CommandHit(f"Jump to PEP{pep.number}", pep.title, GotoPEP(pep.number))


### peps.py ends here



"""Author filtering commands for the command palette."""

##############################################################################
# Textual enhanced imports.
from textual_enhanced.commands import CommandHit, CommandHits, CommandsProvider

##############################################################################
# Local imports.
from ..data import PEPs
from ..messages import ShowAuthor


##############################################################################
class AuthorCommands(CommandsProvider):
    """A command palette provider related to authors."""

    active_peps: PEPs | None = None
    """The currently-active collection of PEPs to get the authors of."""

    @classmethod
    def prompt(cls) -> str:
        """The prompt for the command provider."""
        return (
            "Also search for PEPs authored by..."
            if cls.active_peps and cls.active_peps.is_filtered
            else "Search for PEPs authored by..."
        )

    def commands(self) -> CommandHits:
        """Provide the author-based command data for the command palette.

        Yields:
            The commands for the command palette.
        """
        if self.active_peps is None:
            return
        help_prefix, command_prefix = (
            ("Also filter", "Also authored by")
            if self.active_peps.is_filtered
            else ("Filter", "Authored by")
        )
        for author in self.active_peps.authors:
            yield CommandHit(
                f"{command_prefix} {author.author}",
                f"{help_prefix} to PEPs authored by {author.author} (narrows down to {author.count})",
                ShowAuthor(author.author),
            )


### authors.py ends here



"""Provides a class for handling notes made about PEPs."""

##############################################################################
# Python imports.
from json import dumps, loads
from pathlib import Path
from typing import Final

##############################################################################
# Typing extension imports.
from typing_extensions import Self

##############################################################################
# Local imports.
from .locations import data_dir


##############################################################################
class Notes:
    """Class that holds the notes the user has made about PEPs."""

    _NOTES_FILE: Final[Path] = data_dir() / Path("notes.json")
    """The location of the notes file."""

    def __init__(self) -> None:
        """Initialise the object."""
        self._notes: dict[int, str] = {}
        """The notes."""

    def load(self, source: Path | None = None) -> Self:
        """Load the notes.

        Args:
            source: The optional source location to load the notes from.

        Returns:
            Self.
        """
        if (source := source or self._NOTES_FILE).exists():
            self._notes = {
                int(number): note
                for number, note in loads(source.read_text(encoding="utf-8")).items()
            }
        return self

    def save(self, target: Path | None = None) -> Self:
        """Save the notes.

        Args:
            target: The optional target location to save the notes to.

        Returns:
            Self.
        """
        (target or self._NOTES_FILE).write_text(
            dumps(self._notes, indent=4), encoding="utf-8"
        )
        return self

    def __getitem__(self, pep: int) -> str:
        """Get the note for a PEP."""
        return self._notes.get(pep, "")

    def __setitem__(self, pep: int, notes: str) -> None:
        """Set the note for a PEP."""
        if notes := notes.strip():
            self._notes[pep] = notes
        else:
            del self[pep]

    def __delitem__(self, pep: int) -> None:
        """Remove a note for a PEP."""
        try:
            del self._notes[pep]
        except KeyError:
            # Attempting to delete a note that doesn't exist is fine with
            # us.
            pass

    def __contains__(self, pep: int) -> bool:
        """Are there notes for a PEP?"""
        return bool(self._notes.get(pep))

    def __len__(self) -> int:
        """The count of notes."""
        return len(self._notes)


### notes.py ends here



"""Provides the main application commands for the command palette."""

##############################################################################
# Textual enhanced imports.
from textual_enhanced.commands import CommandHits, CommandsProvider, Help, Quit

##############################################################################
# Local imports.
from ..commands import (
    ChangeTheme,
    EditNotes,
    Escape,
    FindPEP,
    RedownloadPEPs,
    Search,
    SearchAuthor,
    SearchPythonVersion,
    SearchStatus,
    SearchType,
    ShowAll,
    SortByCreated,
    SortByNumber,
    SortByTitle,
    ToggleAuthorsSortOrder,
    TogglePEPDetails,
    TogglePythonVersionsSortOrder,
    ToggleSortOrder,
    ToggleStatusesSortOrder,
    ToggleTypesSortOrder,
    ViewPEP,
)


##############################################################################
class MainCommands(CommandsProvider):
    """Provides some top-level commands for the application."""

    def commands(self) -> CommandHits:
        """Provide the main application commands for the command palette.

        Yields:
            The commands for the command palette.
        """
        yield ChangeTheme()
        yield EditNotes()
        yield Escape()
        yield FindPEP()
        yield Help()
        yield Quit()
        yield RedownloadPEPs()
        yield Search()
        yield SearchAuthor()
        yield SearchPythonVersion()
        yield SearchStatus()
        yield SearchType()
        yield ShowAll()
        yield SortByCreated()
        yield SortByNumber()
        yield SortByTitle()
        yield ToggleAuthorsSortOrder()
        yield TogglePEPDetails()
        yield TogglePythonVersionsSortOrder()
        yield ToggleSortOrder()
        yield ToggleStatusesSortOrder()
        yield ToggleTypesSortOrder()
        yield ViewPEP()


### main.py ends here
